/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.server.notification;

import java.security.GeneralSecurityException;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.funambol.framework.config.Configuration;
import com.funambol.framework.core.bootstrap.BootStrap;
import com.funambol.framework.core.dm.ddf.SyncMLDM;
import com.funambol.framework.logging.FunambolLogger;
import com.funambol.framework.logging.FunambolLoggerName;
import com.funambol.framework.notification.*;
import com.funambol.framework.tools.DbgTools;
import com.funambol.framework.tools.SecurityTools;
import com.funambol.framework.tools.beans.BeanFactory;

import org.apache.commons.lang.StringUtils;

/**
 * Concrete implementation of a <code>NotificationEngine</code>
 *
 * @version $Id: NotificationEngineImpl.java,v 1.1 2006/01/18 18:38:18 luigiafassina Exp $
 */
public class NotificationEngineImpl extends AbstractNotificationEngine {

    // --------------------------------------------------------------- Constants

    private static final float SYNCML_DM_PROTOCOL_VERSION = 1.1f;

    public static final String NOTIFICATION_SENDER_BEAN =
        "com/funambol/server/engine/dm/NotificationSender.xml";

    public static final String BOOTSTRAP_SENDER_BEAN =
        "com/funambol/server/engine/dm/BootstrapSender.xml";


    // -------------------------------------------------------------- Properties

    // ------------------------------------------------------------ Private Data

    private static final Logger log = FunambolLogger.getLogger(FunambolLoggerName.SERVER_DM_NOTIFICATION);

    private Configuration config = null;


    // ------------------------------------------------------------- Constructor

    public NotificationEngineImpl(Configuration config) {
        super(config);
        this.config = config;
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Send a notification message.
     * @param messageType the type of the message. See {@link NotificationConstants} for details.
     * @param transportType the transport type used to send the message. See {@link NotificationConstants} for details.
     * @param sessionId the session id
     * @param phoneNumber the phone number of the target device
     * @param info application specific info
     * @param serverId the server identifier (necessary to calculate the digest of the message)
     * @param serverPassword the server password (necessary to calculate the digest of the message)
     * @param serverNonce the server nonce (necessary to calculate the digest of the message)
     * @throws NotificationException if a error occurs
     */
    public void sendNotificationMessage( int    messageType   ,
                                         int    transportType ,
                                         int    sessionId     ,
                                         String phoneNumber   ,
                                         String info          ,
                                         String serverId      ,
                                         String serverPassword,
                                         byte[] serverNonce
                                       ) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendNotificationMessage with: "+
                     "\n- messageType: "    + messageType    +
                     "\n- transportType: "  + transportType  +
                     "\n- sessionId: "      + sessionId      +
                     "\n- phoneNumber: "    + phoneNumber    +
                     "\n- info: "           + info           +
                     "\n- serverId: "       + serverId       +
                     "\n- serverPassword: " + serverPassword +
                     "\n- serverNonde: "    + serverNonce
                     );
        }

        NotificationMessageBuilder messageBuilder = null;
        Sender messageSender = null;

        messageBuilder = getNotificationMessageBuilder(messageType);

        messageSender = getSender(transportType, messageType);

        if (log.isLoggable(Level.FINE)) {
            log.fine("NotificationMessageBuilder: " + messageBuilder);
            log.fine("Sender: "                     + messageSender );
        }


        // build message to send
        byte[] message = messageBuilder.buildMessage(sessionId, phoneNumber, serverId, serverPassword, serverNonce, SYNCML_DM_PROTOCOL_VERSION);


        if (log.isLoggable(Level.FINE)) {
            log.fine("Notification message to send: " + DbgTools.bytesToHex(message));
        }

        // send message
        messageSender.sendMessage(messageType, phoneNumber, message, info);

        if (log.isLoggable(Level.INFO)) {
            log.info("Notification message to " + phoneNumber + " sent");
        }

    }

    /**
     * Sends more notification messages.
     * @param messageType the type of the message. See {@link NotificationConstants} for details.
     * @param transportType the transport type used to send the message. See {@link NotificationConstants} for details.
     * @param sessionIds the session identifiers
     * @param phoneNumbers the phone numbers list of the target devices
     * @param info application specific info
     * @param serverId the server identifier (necessary to calculate the digest of the message)
     * @param serverPassword the server passwords list (necessary to calculate the digest of the message)
     * @param serverNonce the server nonces list (necessary to calculate the digest of the message)
     * @throws NotificationException if a error occurs
     */
    public void sendNotificationMessages(int      messageType,
                                         int      transportType,
                                         int[]    sessionIds,
                                         String[] phoneNumbers,
                                         String   info,
                                         String   serverId,
                                         String[] serverPasswords,
                                         byte[][] serverNonces) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendNotificationMessages with: "+
                     "\n- messageType: "    + messageType    +
                     "\n- transportType: "  + transportType  +
                     "\n- info: "           + info           +
                     "\n- serverId: "       + serverId       );
        }

        NotificationMessageBuilder messageBuilder = null;
        Sender messageSender = null;

        messageBuilder = getNotificationMessageBuilder(messageType);
        messageSender  = getSender(transportType, messageType);

        if (log.isLoggable(Level.FINE)) {
            log.fine("NotificationMessageBuilder: " + messageBuilder);
            log.fine("Sender: "                     + messageSender );
        }

        int numDevices = phoneNumbers.length;
        byte[][] messages = new byte[numDevices][];
        for (int i=0; i<numDevices; i++) {

            // build message to send
            messages[i] = messageBuilder.buildMessage(sessionIds[i], phoneNumbers[i], serverId,
                serverPasswords[i], serverNonces[i], SYNCML_DM_PROTOCOL_VERSION);

            if (log.isLoggable(Level.FINE)) {
                log.fine("Notification message to send to " + phoneNumbers[i] + ": " +
                         DbgTools.bytesToHex(messages[i]));
            }
        }

        // send message
        messageSender.sendMessages(messageType, phoneNumbers, messages, info);

        if (log.isLoggable(Level.INFO)) {
            log.info("Notification message sent to " + phoneNumbers.length + " devices");
        }

    }



    /**
     * Sends a bootstrap message.
     * @param messageType  the type of the message. See {@link NotificationConstants} for details.
     * @param transportType the transport type used to send the message. See {@link NotificationConstants} for details.
     * @param serverUri the server uri
     * @param deviceUri the device uri
     * @param phoneNumber the phone number of the target device
     * @param syncMLDM the message to send
     * @param info application specific info
     * @throws NotificationException
     */
    public void sendBootstrapMessage(int    messageType,
                                     int    transportType,
                                     String serverUri,
                                     String deviceUri,
                                     String phoneNumber,
                                     SyncMLDM syncMLDM,
                                     String info) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendBootstrapMessage with messageType: " + messageType +
                     ", transportType: " + transportType + ", serverUri: " + serverUri +
                     ", deviceUri: " + deviceUri +
                     ", phoneNumber: " + phoneNumber);
        }

        // verify syncMLDM. Each bootstrap message MUST only contain one dm account
        //if (syncMLDM.getDmAcc().getDMAccounts().size() != 1) {
        //    throw new NotificationException("Bootstrap message MUST only contain one dm account");
        //}

        BootstrapMessageBuilder messageBuilder = null;
        Sender messageSender = null;

        messageBuilder = getBootstrapMessageBuilder(messageType);

        messageSender = getSender(transportType, messageType);

        if (log.isLoggable(Level.FINE)) {
            log.fine("BootstrapMessageBuilder: " + messageBuilder);
            log.fine("Sender: " + messageSender);
        }

        // build message
        byte[] message = messageBuilder.buildMessage(serverUri, deviceUri, phoneNumber, syncMLDM);

        if (log.isLoggable(Level.INFO)) {
            log.info("Bootstrap message to send: " + message + " (length: " + message.length + ")");
        }

        // send message
        messageSender.sendMessage(messageType, phoneNumber, message, info);

        if (log.isLoggable(Level.INFO)) {
            log.info("Bootstrap message to " + phoneNumber + " sent");
        }
    }


    /**
     * Sends bootstrap messages. If in a bootstrap object the digest is null and
     * the imsi isn't null, the digest is computed using HMAC
     * @param messageType  the type of the message. See {@link NotificationConstants} for details.
     * @param transportType the transport type used to send the message. See {@link NotificationConstants} for details.
     * @param bootstrap the bootstrap list
     * @param serverUri the server uri list
     * @param syncMLDM the message list to send
     * @param info application specific info
     * @throws NotificationException
     */
    public void sendBootstrapMessages(int messageType,
                                      int transportType,
                                      BootStrap[] bootstrap,
                                      String serverUri,
                                      SyncMLDM[] syncMLDM,
                                      String info) throws NotificationException {

        if (log.isLoggable(Level.INFO)) {
            log.info("sendBootstrapMessage with messageType: " + messageType +
                     ", transportType: " + transportType + " to " + bootstrap.length + " devices");
        }

        int numDevices = bootstrap.length;

        Sender messageSender = getSender(transportType, messageType);

        byte[][] messages      = new byte[numDevices][];
        String[] macs          = new String[numDevices];
        int[]    authMethods   = new int[numDevices];
        String[] phoneNumbers  = new String[numDevices];

        BootstrapMessageBuilder messageBuilder = null;
        String imsi    = null;
        String userPin = null;

        for (int i=0; i<numDevices; i++) {
            verifyBootStrap(bootstrap[i]);

            phoneNumbers[i] = bootstrap[i].getMsisdn();

            // verify syncMLDM. Each bootstrap message MUST only contain one dm account
            //if (syncMLDM[i].getDmAcc().getDMAccounts().size() != 1) {
            //    throw new NotificationException(
            //        "Bootstrap message MUST only contain one dm account");
            //}

            messageBuilder = getBootstrapMessageBuilder(messageType);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("BootstrapMessageBuilder: " + messageBuilder);
            }

            // build message
            messages[i] = messageBuilder.buildMessage(serverUri,
                                                      bootstrap[i].getDeviceId(),
                                                      bootstrap[i].getMsisdn(),
                                                      syncMLDM[i]);

            authMethods[i] = bootstrap[i].getAuthMethod();

            macs[i] = bootstrap[i].getDigest();
            if (macs[i] == null) {

                imsi = bootstrap[i].getImsi();
                userPin = bootstrap[i].getUserPin();

                if (authMethods[i] == NotificationConstants.AUTH_METHOD_NETWPIN) {

                    if (bootstrap[i].isImsiInSemiOctet()) {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Compute mac using imsi in semi-octect");
                        }

                        macs[i] = computeMac(getKeyFromIMSI(imsi), messages[i]);
                    } else {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Compute mac using imsi in textplain");
                        }

                        macs[i] = computeMac(imsi.getBytes(), messages[i]);
                    }
                } else if (authMethods[i] == NotificationConstants.AUTH_METHOD_USERPIN) {

                    byte[] key = null;
                    //
                    // If userpin is null, I use the imsi also isn't properly correct
                    //
                    if (userPin == null) {

                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Try using USERPIN but userPin is null. Uses imsi to calculate the MAC ");
                        }

                        if (imsi == null) {
                            if (log.isLoggable(Level.FINEST)) {
                                log.finest("Try using USERPIN but userPin and imsi are null");
                            }
                            throw new NotificationException(
                                "Userpin or imsi must be not null using USERPIN authentication method");
                        }

                        if (bootstrap[i].isImsiInSemiOctet()) {

                            if (log.isLoggable(Level.FINEST)) {
                                log.finest("Compute mac using imsi in semi-octect");
                            }

                            key = getKeyFromIMSI(imsi);
                        } else {
                            if (log.isLoggable(Level.FINEST)) {
                                log.finest("Compute mac using imsi in textplain");
                            }

                            key = imsi.getBytes();
                        }
                    } else {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Compute mac using user pin");
                        }
                        key = userPin.getBytes();
                    }

                    macs[i] = computeMac(key, messages[i]);
                } else if (authMethods[i] == NotificationConstants.AUTH_METHOD_USERNETWPIN) {
                    byte[] imsiKey = null;
                    if (bootstrap[i].isImsiInSemiOctet()) {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Compute mac using imsi (in semi-octect) and user pin");
                        }

                        imsiKey = getKeyFromIMSI(imsi);
                    } else {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("Compute mac using imsi (in text-plain) and user pin");
                        }
                        imsiKey = imsi.getBytes();
                    }
                    // Append imsiKey with the user pin
                    int imsiKeyLength = imsiKey.length;
                    byte[] userPinKey = userPin.getBytes();
                    int userPinKeyLength = userPinKey.length;
                    byte[] key = new byte[imsiKeyLength + userPinKeyLength];
                    System.arraycopy(imsiKey, 0, key, 0, imsiKeyLength);
                    System.arraycopy(userPinKey, 0, key, imsiKeyLength, userPinKeyLength);
                    macs[i] = computeMac(key, messages[i]);
                }

            }



            if (log.isLoggable(Level.FINEST)) {
                log.finest("Bootstrap message to send: " + messages[i] + " (length: " + messages[i].length +
                         ")");
            }
        }

        // send message
        messageSender.sendMessages(messageType, CONTENT_TYPE_BOOTSTRAP, macs, authMethods, phoneNumbers, messages, info);
   }

   /**
    * Sends more bootstrap message.
    * @param messageType  the type of the message. See {@link NotificationConstants} for details.
    * @param transportType the transport type used to send the message. See {@link NotificationConstants} for details.
    * @param phoneNumbers the phone numbers
    * @param macs the macs of the messages
    * @param authMethods the authetication methods
    * @param messages the messages
    * @param info application specific info
    * @throws NotificationException if a error occurs
    */
   public void sendBootstrapMessages(int      messageType,
                                     int      transportType,
                                     String[] phoneNumbers,
                                     String[] macs,
                                     int[]    authMethods,
                                     byte[][] messages,
                                     String   info) throws  NotificationException {

       Sender messageSender = null;

       messageSender = getSender(transportType, messageType);

       if (log.isLoggable(Level.FINE)) {
           log.fine("Sender: " + messageSender);
       }

       // send message
       messageSender.sendMessages(messageType, CONTENT_TYPE_BOOTSTRAP, macs, authMethods, phoneNumbers, messages, info);
    }

   /**
    * Preprare bootstrap message
    * @param messageType the type of the message. See {@link NotificationConstants} for details.
    * @param bootstrap the bootstrap list
    * @param serverUri the server uri list
    * @param syncMLDM the message list to send
    * @param info application specific info
    */
   public BootStrap[] prepareBootstrap(int messageType,
                                       BootStrap[] bootstrap,
                                       String serverUri,
                                       SyncMLDM[] syncMLDM,
                                       String info) throws NotificationException {

       if (log.isLoggable(Level.INFO)) {
           log.info("prepareBootstrap with messageType: " + messageType +
                    " for " + bootstrap.length + " devices");
       }

       int numDevices = bootstrap.length;

       byte[][] messages      = new byte[numDevices][];
       String[] digests       = new String[numDevices];
       String[] phoneNumbers  = new String[numDevices];

       BootstrapMessageBuilder messageBuilder = null;

       for (int i=0; i<numDevices; i++) {

           digests[i] = bootstrap[i].getDigest();
           phoneNumbers[i] = bootstrap[i].getMsisdn();

           // verify syncMLDM. Each bootstrap message MUST only contain one dm account
           //if (syncMLDM[i].getDmAcc().getDMAccounts().size() != 1) {
           //    throw new NotificationException(
           //        "Bootstrap message must contain only one dm account");
           //}

           messageBuilder = getBootstrapMessageBuilder(messageType);

           if (log.isLoggable(Level.FINEST)) {
               log.finest("BootstrapMessageBuilder: " + messageBuilder);
           }

           // build message
           messages[i] = messageBuilder.buildMessage(serverUri,
                                                     bootstrap[i].getDeviceId(),
                                                     bootstrap[i].getMsisdn(),
                                                     syncMLDM[i]);

           bootstrap[i].setBootstrapMessage(messages[i]);
           if (log.isLoggable(Level.FINEST)) {
               log.finest(bootstrap[i].toString());
           }
       }

       return bootstrap;
   }



    // --------------------------------------------------------- Private Methods
    /**
     * Returns the bootstrap message builder for the given messageType
     * @param messageType
     * @throws NotificationException
     * @return BootstrapMessageBuilder
     */
    private BootstrapMessageBuilder getBootstrapMessageBuilder(int messageType) throws
        NotificationException {
        BootstrapMessageBuilder messageBuilder = null;

        switch (messageType) {
            case NotificationConstants.MESSAGE_TYPE_BOOTSTRAP_PLAIN:
                messageBuilder = new PlainBootstrapMessageBuilder();
                break;
            case NotificationConstants.MESSAGE_TYPE_BOOTSTRAP_WAP:
                messageBuilder = new WAPBootstrapMessageBuilder();
                break;

            default:
                throw new NotificationException("Message type not supported");
        }

        return messageBuilder;
    }

    /**
     * Returns the sender for the given transportType and messageType.
     * ATTENTION: in this implementation read the sender as bean independently from the transportType value.
     *
     * @param transportType int
     * @throws NotificationException
     * @return Sender
     */
    private Sender getSender(int transportType, int messageType) throws NotificationException {
        Sender messageSender = null;

        // read the sender object from the config path
        ClassLoader classLoader = config.getClassLoader();

        try {
            if (messageType == NotificationConstants.MESSAGE_TYPE_BOOTSTRAP_PLAIN ||
                messageType == NotificationConstants.MESSAGE_TYPE_BOOTSTRAP_WAP) {

                messageSender = (Sender)BeanFactory.getBeanInstance(classLoader, BOOTSTRAP_SENDER_BEAN);

            } else if (messageType == NotificationConstants.MESSAGE_TYPE_NOTIFICATION_GENERIC) {

                messageSender = (Sender)BeanFactory.getBeanInstance(classLoader, NOTIFICATION_SENDER_BEAN);

            } else {
                messageSender = new SimpleSender();
            }

        } catch (Exception ex) {
            log.throwing(getClass().getName(), "getSender", ex);
            throw new NotificationException("Error reading the sender", ex);
        }

        return messageSender;
    }

    /**
     * Returns the notification message builder for the given messageType
     * @param messageType int
     * @throws NotificationException
     * @return NotificationMessageBuilder
     */
    private NotificationMessageBuilder getNotificationMessageBuilder(int messageType) throws
        NotificationException {

        NotificationMessageBuilder messageBuilder = null;

        // get NotificationMessageBuilder
        switch (messageType) {
            case NotificationConstants.MESSAGE_TYPE_NOTIFICATION_GENERIC:
                messageBuilder = new GenericNotificationMessageBuilder();
                break;
            default:
                throw new NotificationException("Message type not supported");
        }

        return messageBuilder;
    }

    /**
     * Compute the mac of the message using the given key.
     *
     * @param key the key to use
     * @param message the message
     * @throws NotificationException if a error occurs
     * @return the mac of the message
     */
    private String computeMac(byte[] key, byte[] message) throws NotificationException {

        byte[] bMac = null;

        try {
            bMac = SecurityTools.computeHmacSha1(key, message);
        } catch (GeneralSecurityException e) {
            throw new NotificationException("Error computing the mac", e);
        }

        String mac  = bytesToHex(bMac);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Mac: " + mac);
        }
        return mac;
    }

    /**
     * Returns hex String representation of given byte[]
     * @param b byte[]
     * @return the hex representation of the byte[]
     */
    private static String bytesToHex(byte[] b) {
        StringBuffer buf = new StringBuffer("");
        for (int i = 0; i < b.length; i++) {
            buf.append(byteToHex(b[i]));
        }
        return buf.toString();
    }

    /**
     * Returns hex String representation of given byte
     * @param b byte
     * @return the hex representation of the byte
     */
    private static String byteToHex(byte b) {
        // Returns hex String representation of byte b
        char hexDigit[] = {
                          '0', '1', '2', '3', '4', '5', '6', '7',
                          '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
        };
        char[] array = {hexDigit[ (b >> 4) & 0x0f], hexDigit[b & 0x0f]};
        return new String(array);
    }

    /**
     * Computes the bytes array to use as key in the mac calculation
     * based on imsi. The IMSI MUST be on semi-octet representation as defined
     * in Digital cellular Telecommunications system (Phase 2+); Specification of the Subscriber
     * Identity Module - Mobile Equipment (SIM - ME) interface
     * (GSM 11.11 version 7.2.0 Release 1998)
     *
     * @param imsi String
     * @return byte[]
     */
    private static byte[] getKeyFromIMSI(String imsi) throws NotificationException {

        imsi = imsi.trim();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("IMSI: " + imsi);
        }

        if ( (imsi.length() % 2) == 1 ) {
            imsi = "9" + imsi;
        } else {
            imsi = "1" + imsi;
            imsi = imsi + "F";

        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("IMSI updated: " + imsi);
        }

        int numDigit = imsi.length();
        String temp = null;
        char c1 = 0;
        char c2 = 0;
        byte b  = 0;
        byte[] key = new byte[numDigit / 2]; // always even
        int t = 0;
        for (int i = 0; i < numDigit; i++) {
            c1 = imsi.charAt(i);
            c2 = imsi.charAt(++i);
            temp = "" + c2 + c1;
            try {
                key[t] = (byte) (Integer.parseInt(temp, 16));
            } catch (NumberFormatException ex) {
                throw new NotificationException("IMSI isn't valid (only numbers are permitted)");
            }
            t++;
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Key from imsi: " + bytesToHex(key));
        }

        return key;
    }


    /**
     * Verify if the given bootstrap is correct (from the semantic point of view)
     * @param bootstrap BootStrap
     * @throws NotificationException if the bootstrap is not correct
     */
    private void verifyBootStrap(BootStrap bootstrap) throws NotificationException {

        if (bootstrap == null) {
            throw new NotificationException(
                "Unable to send the bootstrap (is null)");
        }

        String imsi    = bootstrap.getImsi();
        String userPin = bootstrap.getUserPin();
        int authMethod = bootstrap.getAuthMethod();
        String digest  = bootstrap.getDigest();

        if (StringUtils.isNotEmpty(digest)) {
            return;
        }

        if (authMethod == NotificationConstants.AUTH_METHOD_NETWPIN) {
            if (StringUtils.isEmpty(imsi)) {
                throw new NotificationException(
                    "Imsi can't be null using NETWPIN authentication method");
            }
        } else if (authMethod == NotificationConstants.AUTH_METHOD_USERPIN) {
            if (StringUtils.isEmpty(userPin) && StringUtils.isEmpty(imsi)) {
                throw new NotificationException(
                    "Userpin or imsi must be not null using USERPIN authentication method");
            }
        } else if (authMethod == NotificationConstants.AUTH_METHOD_USERNETWPIN) {
            if (StringUtils.isEmpty(imsi)) {
                throw new NotificationException(
                    "Imsi can't be null using USERNETWPIN authentication method");
            }

            if (StringUtils.isEmpty(userPin)) {
                throw new NotificationException(
                    "Userpin can't be null using USERNETWPIN authentication method");
            }
        }
    }

}
